# -*- coding: utf-8 -*-
"""
Created on Mon Aug 31 10:51:49 2020

@author: Luana Ruiz
"""

"""
LAB 3: RECOMMENDATION SYSTEMS
"""

#\\\ Standard libraries:
import numpy as np
import torch; torch.set_default_dtype(torch.float64)
import torch.nn as nn
import torch.optim as optim
import copy

#\\\ Own libraries:
import data as data
import Modules.architectures as archit
import Utils.graphML as gml

################################
####### DATA GENERATION ########
################################

"""
Load the data, generate the graph and define the training and test sets (items 1.1--1.6)
"""

# Loading and cleaning up the data

X, idxContact = data.load_data(movie=257, min_ratings=150)
N = X.shape[1]

# Creating and sparsifying the graph

nTotal = X.shape[0]
permutation = np.random.permutation(nTotal)
nTrain = int(np.ceil(0.9*nTotal))
idxTrain = permutation[0:nTrain]
nTest = nTotal-nTrain
idxTest=permutation[nTrain:nTotal]

W = data.create_graph(X=X, idxTrain=idxTrain, knn=40)

# Creating the training and test sets

xTrain, yTrain, xTest, yTest = data.split_data(X, idxTrain, idxTest, idxContact)
nTrain = xTrain.shape[0]
nTest = xTest.shape[0]

################################
######## LOSS FUNCTION #########
################################

"""
Write the loss function defined in equation (4) (item 2.1).
"""

def movieMSELoss(yHat,y,idxMovie):
    mse = nn.MSELoss()
    return mse(yHat[:,:,idxMovie].reshape([-1,1]),y[:,:,idxMovie].reshape([-1,1]))

################################
######## ARCHITECTURES #########
################################

architectures = dict()

# Linear parametrization
linearParam = torch.nn.Linear(N,N)
architectures['LinearLayer'] = linearParam

# Fully connected neural network
fcNet = nn.Sequential(torch.nn.Linear(N,64), nn.ReLU(), torch.nn.Linear(64,N), nn.ReLU())
architectures['FCLayer    '] = fcNet

# Graph filter
graphFilter = archit.LocalGNN([1, 64], [5], True, gml.NoActivation, [N],
                              gml.NoPool, [1], [1], W)
architectures['GraphFilter'] = graphFilter

# GNN, 1 layer
GNN1Ly = archit.LocalGNN([1, 64], [5], True, nn.ReLU, [N],
                              gml.NoPool, [1], [1], W)
architectures['GNN 1 layer'] = GNN1Ly

# GNN, 2 layers
GNN2Ly = archit.LocalGNN([1, 64, 32], [5, 5], True, nn.ReLU, [N, N],
                              gml.NoPool, [1, 1], [1], W)
architectures['GNN 2 layer'] = GNN2Ly

################################
########### TRAINING ###########
################################

validationInterval = 30

nEpochs = 40
batchSize = 5
learningRate = 0.005

nValid = int(np.floor(0.1*nTrain))
xValid = xTrain[0:nValid,:,:]
yValid = yTrain[0:nValid,:,:]
xTrain = xTrain[nValid:,:,:]
yTrain = yTrain[nValid:,:,:]
nTrain = xTrain.shape[0]

# Declaring the optimizers for each architectures
optimizers = dict()
for key in architectures.keys():
    optimizers[key] = optim.Adam(architectures[key].parameters(), lr=learningRate)

if nTrain < batchSize:
    nBatches = 1
    batchSize = [nTrain]
elif nTrain % batchSize != 0:
    nBatches = np.ceil(nTrain/batchSize).astype(np.int64)
    batchSize = [batchSize] * nBatches
    while sum(batchSize) != nTrain:
        batchSize[-1] -= 1
else:
    nBatches = np.int(nTrain/batchSize)
    batchSize = [batchSize] * nBatches
batchIndex = np.cumsum(batchSize).tolist()
batchIndex = [0] + batchIndex

epoch = 0 # epoch counter

# Store the training...
lossTrain = dict()
costTrain = dict()
lossValid = dict()
costValid = dict()
# ...and test variables
lossTestBest = dict()
costTestBest = dict()
lossTestLast = dict()
costTestLast = dict()

bestModel = dict()

for key in architectures.keys():
    lossTrain[key] = []
    costTrain[key] = []
    lossValid[key] = []
    costValid[key] = []

while epoch < nEpochs:
    randomPermutation = np.random.permutation(nTrain)
    idxEpoch = [int(i) for i in randomPermutation]
    print("")
    print("Epoch %d" % (epoch+1))

    batch = 0

    while batch < nBatches:
        # Determine batch indices
        thisBatchIndices = idxEpoch[batchIndex[batch]
                                    : batchIndex[batch+1]]

        # Get the samples in this batch
        xTrainBatch = xTrain[thisBatchIndices,:,:]
        yTrainBatch = yTrain[thisBatchIndices,:,:]

        if (epoch * nBatches + batch) % validationInterval == 0:
            print("")
            print("    (E: %2d, B: %3d)" % (epoch+1, batch+1),end = ' ')
            print("")

        for key in architectures.keys():
            # Reset gradients
            architectures[key].zero_grad()

            # Obtain the output of the architectures
            yHatTrainBatch = architectures[key](xTrainBatch)

            # Compute loss
            lossValueTrain = movieMSELoss(yHatTrainBatch, yTrainBatch, idxContact)

            # Compute gradients
            lossValueTrain.backward()

            # Optimize
            optimizers[key].step()

            costValueTrain = np.sqrt(lossValueTrain.item())

            lossTrain[key] += [lossValueTrain.item()]
            costTrain[key] += [costValueTrain]

            # Print:
            if (epoch * nBatches + batch) % validationInterval == 0:
                with torch.no_grad():
                    # Obtain the output of the GNN
                    yHatValid = architectures[key](xValid)

                # Compute loss
                lossValueValid = movieMSELoss(yHatValid, yValid, idxContact)

                # Compute accuracy:
                costValueValid = np.sqrt(lossValueValid.item())

                lossValid[key] += [lossValueValid.item()]
                costValid[key] += [costValueValid]

                print("\t" + key + ": %6.4f / %7.4f [T]" % (
                        costValueTrain,
                        lossValueTrain.item()) + " %6.4f / %7.4f [V]" % (
                        costValueValid,
                        lossValueValid.item()))

                # Saving the best model so far
                if len(costValid[key]) > 1:
                    if costValueValid <= min(costValid[key]):
                        bestModel[key] =  copy.deepcopy(architectures[key])
                else:
                    bestModel[key] =  copy.deepcopy(architectures[key])

        batch+=1

    epoch+=1

print("")

################################
########## EVALUATION ##########
################################

print("Final evaluation results")

for key in architectures.keys():
    with torch.no_grad():
        yHatTest = architectures[key](xTest)
    lossTestLast[key] = movieMSELoss(yHatTest, yTest, idxContact)
    costTestLast[key] = np.sqrt(lossTestLast[key].item())
    with torch.no_grad():
        yHatTest = bestModel[key](xTest)
    lossTestBest[key] = movieMSELoss(yHatTest, yTest, idxContact)
    costTestBest[key] = np.sqrt(lossTestBest[key].item())
    print(" " + key + ": %6.4f [Best]" % (
                        costTestBest[key]) + " %6.4f [Last]" % (
                        costTestLast[key]))

